#!/usr/bin/perl -w

################################################################################
# nmap-auto
#
# Version 1.0
#
# This script uses nmap to scan a list of ips and send an alert for all open
# ports that are listed in the machine profile or an ALARM for open ports that 
# are not listed in the machine profile.
#
# Copyright (c) 2003 University of Utah Student Computing Labs.
# All Rights Reserved.
#
# Permission to use, copy, modify, and distribute this software and
# its documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appears in all copies and
# that both that copyright notice and this permission notice appear
# in supporting documentation, and that the name of The University
# of Utah not be used in advertising or publicity pertaining to
# distribution of the software without specific, written prior
# permission. This software is supplied as is without expressed or
# implied warranties of any kind.
#
################################################################################

################################################################################
# Change these
#
my $basepath = "/Library/StartupItems/Nmap-auto/Resources";  # Required
my $directory_with_nmap = "$basepath/nmap/";                 # Required
my $config_file = "$basepath/nmap-auto.config";              # Required
my $system_logger = "/usr/bin/logger -i -t nmap-auto";       # Save to system log

                                                             # One of the following log methods is requred:
                                                             # (you may use both if desired)
my $remote_logger = "$basepath/report_to_master.sh";         # Option 1 (set to "" or remove script)
my $local_log = "/var/log/nmap_auto.log";                    # Option 2 (set to "" to disable)

################################################################################
# nmap scan settings
#
my $tcp_scantype = "-sS"; # -sS is a TCP SYN scan, which is faster.
my $udp_scantype = "-sU"; # -sU is very slow, change to "" to get rid of it.
my $ports = "-p 1-65535"; # Maximum possible range is 1-65535
my $speed = "-T Normal";  # Faster than Normal will probably be dropped by the client's kernel
                          # -T <Paranoid|Sneaky|Polite|Normal|Aggressive|Insane> 

################################################################################
# Dont change unless you have a good idea what you are doing
#
my $parameters = "-v -n $tcp_scantype $udp_scantype $ports $speed"; # Changing this may make this script not work
my $debug = 0; # Change to 1 and uncomment the if ($debug)... lines to get verbose
my $file_open_error = 0; # Don't change this at all
my $pid_file = "/var/run/nmap-auto.pid";

################################################################################
# Go
#

##
# Force root user (to save pid and to run certain nmap scans)
if ($< != 0 ) {
  system "$system_logger \"You must be root to run this script\"";
  die "You must be root to run this script\n";
}

##
# Sanity check
if (! -d "$basepath") {                                                           # Base directory
  system "$system_logger \"Not found: $basepath\"";
  die "Not found: $basepath\n"; 
}
if (! -f "$directory_with_nmap/nmap") {                                           # Nmap
  system "$system_logger \"Not found: $directory_with_nmap/nmap\"";
  die "Not found: $directory_with_nmap/nmap\n"; 
}
if (! -f "$config_file") {                                                        # Config file
  system "$system_logger \"Not found: $config_file\"";
  die "Not found: $config_file\n"; 
}
if (! -f "$remote_logger" && "$local_log" eq "" ) {                                # Remote logger script
  system "$system_logger \"Not found: $remote_logger or $local_log\"";
  die "Not found: $remote_logger or $local_log\n";
}

##
# Catch some kills
$SIG{'TERM'} = \&sig_handler;
$SIG{'INT'} = \&sig_handler;

##
# Save PID
if ( -e $pid_file ) {
  unlink $pid_file;
}
open (FILEHANDLE, "> $pid_file") or $file_open_error = 1;
if ($file_open_error) {
  system "$system_logger \"Could not open $pid_file\"";
  die "Could not open $pid_file\n"; 
}
print FILEHANDLE "$$\n";
close (FILEHANDLE);

## Loop forever
while (1) {

  my $name_of_profile;
  my @config_file_contents;
  
  ##
  # Get list of IP's to scan
  open (FILEHANDLE, "< $config_file") or $file_open_error = 1;
  if ($file_open_error) {
    system "$system_logger \"Could not open $config_file\"";
    unlink $pid_file;  # Remove PID file
    die "Could not open $config_file\n"; 
  }
  @config_file_contents = <FILEHANDLE>;
  close (FILEHANDLE);

  ##
  # Parse config file
  $number_of_lines = @config_file_contents;
  for ( $i = 0; $i < $number_of_lines; $i++ ) {
    my $line = $config_file_contents[$i];
    chomp $line;
    if ($line ne "" && substr($line, 0, 1) ne '#') {                                #### Find lines that contain box profiles
      if (substr($line, 0, 1) eq '@') {             
        ($name_of_profile) = split (' ', $line);
        my @tempy = ();
        $i++;
        $line = $config_file_contents[$i];
        while (index($line, '}') < 0 && $i < $number_of_lines) {
          if ($line ne "" && substr($line, 0, 1) ne '#') {
            @tempy2 = split (' ', $line);
            push @tempy, [@tempy2];
          }
          $i++;
          $line = $config_file_contents[$i];
        }
        $type_of_box{$name_of_profile} = [ @tempy ];
  #      if ($debug) {for $num2 ( @tempy ) { print "[ @$num2 ],\n"; } }
      } else {                                                                      #### Add IP
        push(@ip_list,[$line]);
      }
    }
  }

  ##
  # Scan the IP's one by one
  for $this_ip_list ( @ip_list ) {
    for $nummb (@$this_ip_list) {
      my @tempy  = split (' ', $nummb);
      $ip = $tempy[0];
      $this_type_of_box = $tempy[1];
    }
    @known_ports = @{ $type_of_box{$this_type_of_box} };
    if (index ($ip, '-') > 0) {                                                     #### Found an IP range
      @ip_parts = split ('\.', $ip);
      @ip_range = split ('-', $ip_parts[3]);
      for ($loopity = $ip_range[0]; $loopity <= $ip_range[1]; $loopity++) {
        $theip = "$ip_parts[0].$ip_parts[1].$ip_parts[2].$loopity";
  #      if ($debug) {print "$theip $this_type_of_box\n"};
  #      if ($debug) {for $num2 ( @known_ports ) { print "[ @$num2 ],\n"; } }
        &scan_ip($theip, @known_ports);                                             #### Scan IP
      }
    } else {                                                                        #### Found an IP range
  #    if ($debug) {print "$ip $this_type_of_box\n"};
  #    if ($debug) {for $num2 ( @known_ports ) { print "[ @$num2 ],\n"; } }
      &scan_ip($ip, @known_ports);                                                  #### Scan IP
    }
  }
}

sub scan_ip {
  ################################################
  # scan_ip v1
  ################################################

  my ($ip, @known_ports) = @_;
  my @output1;
  my $output2;
  my @output3;
  my $i;
  my $j;
  my @tempy_ports;
  my @ports_hash;
  my $scan_error = 0;

  my $all_ports;
  my @alarm_notification;
  my @alert_notification;
  my $total_number_of_ports = 0;
  my $number_of_filtered_ports = 0;
  my $number_of_alarm_ports = 0;
  my $number_of_alert_ports = 0;
  my $status;
  my $logger_message;
  my $alarm;
  my $down;
  my $port_number;
  my $port_state;
  my $tcp_or_udp;
  my $service_name;

  # Scan the IP
  system "$system_logger \"Executing: nmap $parameters -oG - $ip\"";
  if ($debug) {print "Executing: nmap $parameters -oG - $ip\n";}
  chdir $directory_with_nmap;
  $output2 = `./nmap $parameters -oG - $ip`;
  if ($debug) {print "$ip done, parsing\n";}
#  if ($debug) {print "Output:\n$output2";}

  # parse output
  @output3 = split ('\n', $output2);
  foreach $i (@output3) {
    if (substr ($i, 0, 4) eq "Host") {
      $down = 0;
      $scan_error = 0;
      @output1 = split (/\t/ , $i);
      $all_ports = $output1[1];
      @tempy_ports = split (/ / , $all_ports);
      shift @tempy_ports;
      foreach $j (@tempy_ports) {
        if ($debug) {print "$j\n";}
        
        if ("$j" eq "Down") {
          $down = 1;
          last;
        } elsif ("$j" eq "Up") {
          $scan_error = 1;
          last;
        } else {
          push (@ports_hash, $j);
        }
      }
      last;
    }
  }

#  if ($debug) {print "ports_hash @ports_hash\n";}

  # Compare ports lists.
  $total_number_of_ports = @ports_hash;
  $number_of_known_ports = @known_ports;

  if (! $scan_error && ! $down) {
    for ($i = 0; $i < $total_number_of_ports; $i++) {
      $alarm = 1;
      $filtered = 0;
      @port_scan_parts = split (/\// , $ports_hash[$i]);

      # Examine each port message
      $port_number = $port_scan_parts[0];
      $port_state = $port_scan_parts[1];
      $tcp_or_udp = $port_scan_parts[2];
      $service_name = $port_scan_parts[4];
              
      if ($port_state eq "filtered") {
        $alarm = 0;
        $filtered = 1;
      } else {
        for ($j = 0; $j < $number_of_known_ports; $j++) {
          if (index ($known_ports[$j][0], '-') >= 0) {
            @port_range = split ('-', $known_ports[$j][0]);
  #          if ($debug) {$blabla = index ($known_ports[$j][0], '-');}
  #          if ($debug) {print "index '-' $blabla j $j Known ports $known_ports[$j][0] port_range @port_range\n";}
            $number = @port_range;
            if ($number == 2) {
              if ($port_number >= $port_range[0] && $port_number <= $port_range[1] && $known_ports[$j][1] eq $tcp_or_udp ) {
                $alarm = 0;
                # IP is between range, match!
                last;
              } 
            } else {
              system "$system_logger \"Port range error: $known_ports[$j][0]\""; # don't understand the line.
              unlink $pid_file;  # Remove PID file
              die "Port range error: $known_ports[$j][0]";
            }
          } elsif($known_ports[$j][0] == $port_number && $known_ports[$j][1] eq $tcp_or_udp) {
            $alarm = 0;
            last; #$j = $number_of_known_ports; # end loop
          }
        }
      }

      if ($filtered) {
        $number_of_filtered_ports++;
        if ($debug) {print "Filtered: $port_number $port_state $tcp_or_udp $service_name, \n";}
      } elsif ($alarm) {
        push(@alarm_notification,"$port_number $port_state $tcp_or_udp $service_name,");
        $number_of_alarm_ports++;
        if ($debug) {print "alarm $port_number $port_state $tcp_or_udp $service_name, \n";}
      } else {
        push(@alert_notification,"$port_number $port_state $tcp_or_udp $service_name,");
        $number_of_alert_ports++;
        if ($debug) {print "alert $port_number $port_state $tcp_or_udp $service_name, \n";}
      }
    }

    if ($debug) {
      $debug1 = @alarm_notification;
      $debug2 = @alert_notification;
      print "alarms $debug1, alerts $debug2\n";
      foreach $k (@alarm_notification) {print "alarm_notification $k\n";} 
      foreach $k (@alert_notification) {print "alert_notification $k\n";} 
    } 

    ## 
    # Report
    $status = "Alert";
    $logger_message = "$ip, total: $total_number_of_ports, filtered: $number_of_filtered_ports, alarm: $number_of_alarm_ports, alert: $number_of_alert_ports.";
    if ($number_of_alarm_ports != 0) {
      $logger_message .= " ALARM: @alarm_notification";
      $status = "ALARM";
    }
    if ($number_of_alarm_ports != 0) {
      $logger_message .= " Alert: @alert_notification";
    }

  } elsif ($down) {
    $logger_message .= "$ip, Down";
    $status = "DOWN";
  } else {
    $logger_message .= "$ip, no ports open...";
    $status = "ERROR";
  }

  nmap_report("NMAP-$status", "$logger_message");

}

sub nmap_report {
  ################################################
  # nmap_report v1
  ################################################

  my ($subject, $message) = @_;

  ##
  # Remote logger
  if ( -x $remote_logger ) {
    system "$remote_logger \"$subject\" \"$message\"";
  }

  ##
  # Local log file
  if ( "$local_log" ne "" ) {
    open (FILEHANDLE, ">> $local_log") or $file_open_error = 1;
    if ($file_open_error) {
      system "$system_logger \"Could not open $local_log\"";
      unlink $pid_file;  # Remove PID file
      die "Could not open $local_log\n"; 
    }
    print FILEHANDLE "$subject $message\n";
    close (FILEHANDLE);
  }
}

sub sig_handler {
    unlink $pid_file;  # Remove PID file
    exit( 0 );
}